#version 330

#define LIGHTS 16

// MATERIAL PROPERTIES ----

uniform bool hasDiffuseMap;
uniform sampler2D texDiffuseMap;

uniform bool hasNormalMap;
uniform sampler2D texNormalMap;

uniform bool hasSpecularMap;
uniform sampler2D texSpecularMap;

uniform bool hasShadowMap;
uniform sampler2D texShadowMap;

uniform bool hasCubeMap;
uniform samplerCube texCubeMap;

uniform vec4 vMaterialAmbient;
uniform vec4 vMaterialDiffuse;
uniform vec4 vMaterialEmissive; 
uniform vec4 vMaterialSpecular;
uniform vec4 vMaterialReflect;
uniform float sMaterialShininess;

// ------------------------

// LIGHTS ----------------

uniform vec4 vLightAmbient;

uniform int iLightType[LIGHTS];

// the light direction is the vector which looks towards the light
uniform vec4 vLightDir[LIGHTS];
uniform vec4 vLightPos[LIGHTS];
uniform vec4 vLightAtt[LIGHTS];
uniform vec4 vLightSpot[LIGHTS]; //(outer angle , inner angle, falloff, free)
uniform vec4 vLightColor[LIGHTS];

// -------------------------

uniform vec3 cameraPos;
uniform vec3 directionalLight;

in vec4 vWorldPos;
in vec4 vColor;
in vec3 vNormal;
in vec3 vTangent;
in vec3 vBitangent;
in vec2 vUv;

out vec4 fColor;

vec3 calcNormal(){
	vec3 tempNormal = vNormal;

	if(hasNormalMap){
		tempNormal = normalize(tempNormal);
		vec3 tempTangent = normalize(vTangent);
		vec3 tempBitangent = normalize(vBitangent);
		
		vec4 normal = texture(texNormalMap, vUv);
		normal = (normal * 2.0) - 1.0;
		tempNormal = tempNormal + normal.x * tempTangent + normal.y * tempBitangent;
	}
	
	return normalize(tempNormal);
}

//--------------------------------------------------------------------------------------
// Blinn-Phong Lighting Reflection Model
//--------------------------------------------------------------------------------------
// Returns the sum of the diffuse and specular terms in the Blinn-Phong reflection model.
vec4 calcBlinnPhongLighting( vec4 LColor, vec4 vMaterialTexture, vec3 N, vec3 L, vec3 H )
{

	vec4 Id = vMaterialTexture * vMaterialDiffuse * clamp( dot(N,L), 0.0f, 1.0f );
	vec4 Is = vMaterialSpecular * pow( clamp(dot(N,H), 0.0f, 1.0f), sMaterialShininess );
	return (Id + Is) * LColor;
	
}

vec4 cubeMapReflection( vec3 V, vec3 N, vec4 I )
{
	vec3 r = reflect( -V, N );

	return mix( I, I * texture(texCubeMap, r), /*specularStrength * */vMaterialReflect.r );

}

//float shadowStrength( vec2 shadowCoords ) { return 1.0f; }

void main() {
	vec3 N = calcNormal();
	vec3 V = normalize(cameraPos - vWorldPos.xyz); 
	
	float s = 1;
	
	//if(hasShadowMap)
	//{
	//	s = shadowStrength( vec2(0) );
	//}
	
	vec4 vMaterialTexture = vec4(1.0);

	if(hasDiffuseMap)
	{			
		// SamplerState is defined in Common.fx.
		vMaterialTexture *= texture(texDiffuseMap, vUv);
	}

	vec4 I = vMaterialEmissive + vMaterialAmbient * vMaterialTexture * vLightAmbient;

	for (int i = 0; i < LIGHTS; i++)
	{		
		if(iLightType[i] == 1) // directional
		{
			vec3 d = normalize( vLightDir[i].xyz );  // light dir	
			vec3 h = normalize( V + d );
			I += s * calcBlinnPhongLighting( vLightColor[i], vMaterialTexture, N, d, h );			
		}
		else if(iLightType[i] == 2)  // point
		{
			vec3 d = ( vLightPos[i] - vWorldPos ).xyz;	 // light dir
			float dl = length(d);
			d = normalize(d);	
			vec3 h = normalize( V + d );
			float att = 1.0f / ( vLightAtt[i].x + vLightAtt[i].y * dl + vLightAtt[i].z * dl * dl );
			I += att*calcBlinnPhongLighting( vLightColor[i], vMaterialTexture, N, d, h );
		}
		else if(iLightType[i] == 3)  // spot
		{			
			vec3 d = ( vLightPos[i] - vWorldPos ).xyz;	// light dir
			float  dl = length(d);							// light distance
			d = normalize(d);								// normalized light dir					
			vec3 h = normalize( V + d );				// half direction for specular
			vec3 sd = normalize( vLightDir[i].xyz );	// missuse the vLightDir variable for spot-dir
		
			float rho = dot(-d, sd);
			float spot = pow( clamp( (rho - vLightSpot[i].x) / (vLightSpot[i].y - vLightSpot[i].x), 0.0f, 1.0f ), vLightSpot[i].z );
			float att = spot / ( vLightAtt[i].x + vLightAtt[i].y * dl + vLightAtt[i].z * dl * dl );
			I += att*calcBlinnPhongLighting( vLightColor[i], vMaterialTexture, N, d, h );
		}
	}		
	I.a = vMaterialDiffuse.a;

	if(hasCubeMap)
	{	
		I = cubeMapReflection( V, N, I );
	}

	fColor = I;
}